//  Template Tagger Package
//  Version 0.3.0 (December 1, 2012)
//  Questions/Comments?  jose.juknevich@weizmann.ac.il

// Copyright (c) 2011-12, Jose Juknevich, Mihailo Backovic, and Raz Alon.
//
//----------------------------------------------------------------------
// This file is part of the Template Tagger package ("Template Overlap").
//
//  Template Tagger is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//----------------------------------------------------------------------


#ifndef __TEMPLATE_TAGGER_HH__
#define __TEMPLATE_TAGGER_HH__

#include "matching.hh"

#include "fastjet/FunctionOfPseudoJet.hh"
#include <fastjet/Selector.hh>
#include <fastjet/CompositeJetStructure.hh> // to derive the FilterStructure from CompositeJetStructure
#include <fastjet/tools/Transformer.hh>     // to derive Filter from Transformer



#include <string>
#include <climits>
#include <utility>

static const double TINY = 1e-20;


using namespace TemplateOverlap;


FASTJET_BEGIN_NAMESPACE      // defined in fastjet/internal/base.hh


/// Forward declarations for TemplateTagger and TemplateTaggerStructure
class TemplateTagger;
class TemplateTaggerStructure;

class MultiTemplateTagger;
class MultiTemplateTaggerStructure;

class LepTopTemplateTagger;
class LepTopTemplateTaggerStructure;

/// This class wraps the core Template Tagger code to provide the fastjet::FunctionOfPseudoJet
/// interface for convenience in larger analyses.  See matching.hh for
/// definitions of Ov_N and the constructor options.

class Noverlap : public FunctionOfPseudoJet<templ_t> {
public:

  Noverlap(double subConeRadius, double sigma,  const string & templateFile, Resolution_scale_scheme variableCone, Jet_shape_scheme mode)
  :  _templateFinder(templateFile, TemplateDefinition(subConeRadius, sigma,variableCone, mode))
{_number_of_templates=_templateFinder.get_number_of_templates();
  std::cout << "Using " << _number_of_templates <<" templates"<<std::endl;
}
   
   /// returns Ov_N, measured on the constituents of this jet
  templ_t result(const PseudoJet& jet) const {return _templateFinder.getOv(jet);}

private:

   mutable TemplateOverlap::MatchingMethod _templateFinder; 
   int _number_of_templates;
};




class planarflow : public FunctionOfPseudoJet<double> {
public:

  planarflow(double r1, double r2, double r3): _r1(r1), _r2(r2), _r3(r3) {}
   
   /// returns Pf, measured on the constituents of this jet
  double result(const PseudoJet& jet) const {
  double Ixy = 0.,Ixx = 0.,Iyy = 0.;
  double etaJet = jet.rapidity();
  double phiJet = jet.phi();
  
  std::vector<fastjet::PseudoJet> particles = jet.constituents();

  double r[3] = {_r1,_r2, _r3};
  for ( unsigned int i = 0; i < particles.size(); ++i) {
    //Get x and y components in that basis.
    double  etaNow = particles[i].rapidity() - etaJet , phiNow = particles[i].phi() - phiJet,
            ee = particles[i].perp();
    if (abs(phiNow) > M_PI) phiNow = 2 * M_PI - abs(phiNow);
    // Calculate matrix to be used in the computation of planar flow
    
    Ixx += ee * (etaNow  * etaNow + r[i] * r[i]/4) ;  
    Iyy += ee * (phiNow * phiNow + r[i] * r[i]/4) ;
    Ixy += ee * etaNow * phiNow; 
    }         

    //Return planar flow.
    return(4 * (Ixx * Iyy -Ixy * Ixy)/((Ixx + Iyy )*(Ixx + Iyy )) );
    
  }

private:
   double _r1, _r2, _r3;
   
};

class angularity : public FunctionOfPseudoJet<double> {
public:

  angularity(int N): _N(N) {}
   
  double result(const PseudoJet& jet) const {
     std::vector<PseudoJet> particles = jet.constituents();
    double sum = 0.;
    for (unsigned int i = 0; i < particles.size(); ++i)
    {
      double Theta = max (sqrt(particles[i].plain_distance(jet)), TINY);
      double cosTheta = cos(Theta);
      double sinTheta = sqrt(1 - cosTheta * cosTheta);
      sum+= particles[i].perp() * pow(sinTheta, _N) * pow(1-cosTheta, 1-_N);
    }
    return (1/jet.m() * sum);
  }

private:
   int _N;
   
};

class thetaBar : public FunctionOfPseudoJet<double> {
public:

  thetaBar(int N): _N(N) {}
   
  double result(const PseudoJet& jet) const {
    std::vector<PseudoJet> particles = jet.constituents();
    double sumTheta = 0.;
    for (unsigned int i = 0; i < particles.size();++i) 
      sumTheta += sin( max (sqrt(particles[i].plain_distance(jet)), TINY));
	
    return sumTheta;
  }

private:
   int _N;
   
};

class thetaS : public FunctionOfPseudoJet<double> {
public:

  thetaS() {}
   
  double result(const PseudoJet& jet) const {
    std::vector<PseudoJet> particles = jet.constituents();
    int iMin  = 0;
    double pTmin = particles[0].perp();
    for (unsigned int i = 1; i < particles.size();++i) 
    { double pTnow = particles[i].perp();
      if (pTnow < pTmin) {iMin = i; pTmin = pTnow;}
    }
    return sqrt(particles[iMin].plain_distance(jet));
  }
   
};



//---------------------------------------------------------------------------------
// A new implementation of the Template Tagger that does not make use of Noverlap
//---------------------------------------------------------------------------------

class TemplateTagger : public Transformer {
public:
  /// ctor with arguments (see the class description above)
   TemplateTagger(const string & templateFile, const TemplateDefinition & templDef , Selector selector ) : 
    _templateFinder(templateFile, templDef), _selector(selector), 
    _subConeRadius(templDef.r()), _ovcut(templDef.ovcut()), _mode(templDef.mode())
    {_number_of_templates=_templateFinder.get_number_of_templates();
  std::cout << "Using " << _number_of_templates <<" templates"<<std::endl;}
    
  /// returns a textual description of the tagger 
  virtual std::string description() const;

  /// runs the tagger on the given jet and
  /// returns the tagged PseudoJet if successful, a PseudoJet==0 otherwise
  /// (standard access is through operator()).
  virtual PseudoJet result(const PseudoJet & jet) const;

  /// the type of Structure returned
  typedef TemplateTaggerStructure StructureType;

protected:
  
  mutable TemplateOverlap::MatchingMethod _templateFinder; 
  mutable Selector _selector;  /// the subjet selection criterium
  double _subConeRadius; /// The size of the subcone around each template parton
  double _ovcut;                 /// the minimum overlap value
  double _sigma;
  int _number_of_templates;
  Jet_shape_scheme _mode; 
  
   
};

class TemplateTaggerStructure : public CompositeJetStructure{
public:
  /// ctor with pieces initialisation
  TemplateTaggerStructure(const std::vector<PseudoJet> & pieces_in) :
    CompositeJetStructure(pieces_in), _ov(0.0){}

  /// returns the associated N-overlap
  inline double ov() const{return _ov;}
  inline std::vector<PseudoJet> maxTempl() const {return _peak_template;}
  
  virtual std::string description() const { return "Template Tagger"; }


protected:
  double _ov;      ///< the value of the N-overlap
  std::vector<PseudoJet> _peak_template; ///< the best matched template

  // allow the tagger to set these
  friend class TemplateTagger;
};

//----------------------------------------------------------------------
// TemplateTagger class implementation
//----------------------------------------------------------------------

// class description
string TemplateTagger::description() const {
  ostringstream ostr;
  ostr << "Filter with subjet_def = ";
  if (_number_of_templates>0)
    ostr << "Template Tagger with cone";
  ostr<< ", selection " << _selector.description();
  
  return ostr.str();
}


//----------------------------------------------------------------------

// return a vector of subjets, which are the ones that would be kept
// by the filtering
PseudoJet TemplateTagger::result(const PseudoJet &jet) const {
  templ_t my_result =  _templateFinder.getOv(jet);
  if (my_result.first < _ovcut || ! _selector.pass(jet) ) return PseudoJet();
  
    std::vector<fastjet::PseudoJet> subjets;
    fastjet::Selector sel = fastjet::SelectorCircle(_subConeRadius);


  if (_mode == TOP)
  {
    for (unsigned int i=0; i< my_result.second.size(); ++i)
    {
     // cout << "testing result() #" <<i <<endl;
      sel.set_reference(my_result.second[i]);
      if (jet.pieces()[i].has_constituents())
      {
	std::vector<fastjet::PseudoJet> particles =  jet.pieces()[i].constituents();
	subjets.push_back(fastjet::join(sel(particles)));
      }
      else
	subjets.push_back(jet.pieces()[i]);
    }
 
  }
  else
  {
    std::vector<fastjet::PseudoJet> particles =  jet.constituents();
    for (unsigned int i=0; i< my_result.second.size(); ++i)
    {
      sel.set_reference(my_result.second[i]);
      fastjet::PseudoJet subjet = fastjet::join(sel(particles));
      subjets.push_back(subjet);
    }
  }
  
  PseudoJet result = join<StructureType>(subjets);
  StructureType * s = (StructureType *) result.structure_non_const_ptr();
  s->_ov = my_result.first;
  s->_peak_template = my_result.second;
  return result;
}



//---------------------------------------------------------------------------------


//---------------------------------------------------------------------------------
// A new implementation of the Template Tagger that reads several template libraries
//---------------------------------------------------------------------------------


class MultiTemplateTagger : public Transformer {
public:
  /// ctor with arguments (see the class description above)
   MultiTemplateTagger(std::vector<FunctionOfPseudoJet<templ_t> *> Ov_func_vec, Selector selector,
		     const double R1, const double ovcut, const std::vector<double> & list_of_pt ) : 
    _Ovfunc_vec(Ov_func_vec), _selector(selector), _R1(R1), _ovcut(ovcut) {_list_of_pt = list_of_pt;}
    
  /// returns a textual description of the tagger 
  virtual std::string description() const;

  /// runs the tagger on the given jet and
  /// returns the tagged PseudoJet if successful, a PseudoJet==0 otherwise
  /// (standard access is through operator()).
  virtual PseudoJet result(const PseudoJet & jet) const;

  /// the type of Structure returned
  typedef MultiTemplateTaggerStructure StructureType;
  
private:
  int _select_template_pt(const double jetPt) const;
  std::vector <double> _list_of_pt; 

protected:
  
  std::vector<FunctionOfPseudoJet<templ_t> *>_Ovfunc_vec; /// the overlap function to use to match the jets
  mutable Selector _selector;  /// the subjet selection criterium
  double _R1; /// The subcone radius
  double _ovcut;                 /// the minimum overlap value
   
};

class MultiTemplateTaggerStructure : public CompositeJetStructure{
public:
  /// ctor with pieces initialisation
  MultiTemplateTaggerStructure(const std::vector<PseudoJet> & pieces_in) :
    CompositeJetStructure(pieces_in), _ov(0.0){}

  /// returns the associated N-overlap
  inline double ov() const{return _ov;}
  inline std::vector<PseudoJet> maxTempl() const {return _peak_template;}
  
  virtual std::string description() const { return "Template Tagger"; }


protected:
  double _ov;      ///< the value of the N-overlap
  std::vector<PseudoJet> _peak_template; ///< the best matched template

  // allow the tagger to set these
  friend class MultiTemplateTagger;
};

//----------------------------------------------------------------------
// TemplateTagger class implementation
//----------------------------------------------------------------------

// class description
string MultiTemplateTagger::description() const {
  ostringstream ostr;
  ostr << "Filter with subjet_def = ";
  if (_Ovfunc_vec.size()>0)
    ostr << "Template Tagger with cone";
  ostr<< ", selection " << _selector.description();
  
  return ostr.str();
}


//----------------------------------------------------------------------

// return a vector of subjets, which are the ones that would be kept
// by the filtering
PseudoJet MultiTemplateTagger::result(const PseudoJet &jet) const {
  double jetPt = jet.perp();
  int iTemp = _select_template_pt(jetPt);
  
  std::cout<< "Using template library #"<<iTemp+1 <<" with pT="<<_list_of_pt[iTemp] << std::endl;
  templ_t my_result =  (*(_Ovfunc_vec[iTemp]))(jet);
  if (my_result.first < _ovcut || ! _selector.pass(jet) ) return PseudoJet();
     
  std::vector<fastjet::PseudoJet> particles =  jet.constituents();
  fastjet::Selector sel = fastjet::SelectorCircle(_R1);
  std::vector<fastjet::PseudoJet> subjets;
  for (unsigned int i=0; i< my_result.second.size(); ++i){
    sel.set_reference(my_result.second[i]);
    fastjet::PseudoJet subjet = fastjet::join(sel(particles));
    subjets.push_back(subjet);
  }
  PseudoJet result = join<StructureType>(subjets);
  StructureType * s = (StructureType *) result.structure_non_const_ptr();
  s->_ov = my_result.first;
  s->_peak_template = my_result.second;
  return result;
}
//Select template library based on jet pT/E
int MultiTemplateTagger::_select_template_pt(const double jetPt) const
{
  if (_list_of_pt.size() == 0) return -1;
  if (_list_of_pt.size() == 1) return 0;
  double deltaPtMin = abs(_list_of_pt[0] - jetPt);
  int iTemp = 0;
  for (unsigned int i=1; i < _list_of_pt.size(); ++i)
  {
    double deltaPt = abs(_list_of_pt[i] - jetPt);
    if (deltaPt < deltaPtMin) {deltaPtMin = deltaPt; iTemp = i;}    
  }
  return iTemp;
}

FASTJET_END_NAMESPACE      // defined in fastjet/internal/base.hh

#endif  // __TEMPLATE_TAGGER_HH__
